/***************************************************************************
 *   Copyright (C) 2006 by Borko Boskovic                                  *
 *   borko.boskovic@uni-mb.si                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "init.h"
#include "piece.h"
#include "square.h"
#include "position.h"
#include "bit.h"
#include "search.h"
#include "threadSearch.h"
#include "uci.h"

const gint Pawn_File[8] = {-3, -1, +0, +1, +1, +0, -1, -3};
const gint Pawn_Rank[8] = {+0, +0, +0, +0, +2, +4, +8, +0};
const gint Knight_Line[8] = {-4, -2, +0, +1, +1, +0, -2, -4};
const gint Knight_Rank[8] = {-2, -1, +0, +1, +2, +3, +2, +1};
const gint Bishop_Line[8] = {-3, -1, +0, +1, +1, +0, -1, -3};
const gint Rook_File[8] = {-2, -1, +0, +1, +1, +0, -1, -2};
const gint Queen_Line[8] = {-3, -1, +0, +1, +1, +0, -1, -3};
const gint King_File[8] = {+3, +4, +2, +0, +0, +2, +4, +3};
const gint King_Rank[8] = {+1, +0, -2, -3, -4, -5, -6, -7};
const gint King_Line[8] = {-3, -1, +0, +1, +1, +0, -1, -3};
const gint Pawn_File_Opening = 5;
const gint Pawn_Rank_Endgame = 2;
const gint Knight_Center_Opening = 5;
const gint Knight_Rank_Opening = 5;
const gint Knight_Center_Endgame = 5;
const gint Knight_Back_Rank_Opening = 0;
const gint Knight_Trapped = 100;
const gint Bishop_Center_Opening = 2;
const gint Bishop_Center_Endgame = 3;
const gint Bishop_Back_Rank_Opening = 10;
const gint Bishop_Diagonal_Opening = 4;
const gint Rook_File_Opening = 3;
const gint Queen_Center_Opening = 0;
const gint Queen_Center_Endgame = 4;
const gint Queen_Back_Rank_Opening = 5;
const gint King_File_Opening = 10;
const gint King_Rank_Opening = 10;
const gint King_Center_Endgame = 12;

guint64 squareKey[14][64];
guint64 enPassantKey[8];
guint64 castleKey[4];
guint64 turnKey;
gint pvt[14][64][2];
guint64 attackPawn[2][64];
guint64 attackKnight[64];
guint64 attackKing[64];
guint64 kingSafeDistance1[64];
guint64 kingSafeDistance2[64];
guint64 maskKnight[64];
guint64 maskBishop[64];
guint64 maskRook[64];
guint64 maskQueen[64];
guint8 bitsIn16bits[0x1u << 16];

void initZobrist(GRand* rand);
void initPvt(void);
void initAttack(void);
void initBitsIn16bits(void);
void initUCIOptions(void);
void initmagicmoves(void);

void init(GRand* mrand){
	initBitsIn16bits();
	initZobrist(mrand);
	initmagicmoves();
	initAttack();
	initPvt();
}

void initZobrist(GRand* mrand){
    gint i,j;
    for(i=0; i<14; i++)
        for(j=0; j<64; j++)
            squareKey[i][j] = g_rand_int(mrand) | ((gint64)g_rand_int(mrand) << 32);

	for(i=0; i<8; i++)
            enPassantKey[i] = g_rand_int(mrand) | ((gint64)g_rand_int(mrand) << 32);
	
	for(i=0; i<4; i++)
            castleKey[i] = g_rand_int(mrand) | ((gint64)g_rand_int(mrand) << 32);

	turnKey = g_rand_int(mrand) | ((gint64)g_rand_int(mrand) << 32);
}

void initPvt(void){
	gint i,j,k;
	for(i=0; i<14; i++){
		for(j=0; j<64; j++){
			for(k=0; k<2; k++){
				pvt[i][j][k] = 0;
			}
		}
	}

	/* Pawns */
	for(i=0; i<64; i++){
		pvt[Pawn][i][Opening] += Pawn_File[file(i)] * Pawn_File_Opening;
		pvt[Pawn][i][Endgame] += Pawn_Rank[rank(i)] * Pawn_Rank_Endgame;
	}

	/* Center control */
	pvt[Pawn][d3][Opening] += 10;
	pvt[Pawn][e3][Opening] += 10;
	pvt[Pawn][d4][Opening] += 20;
	pvt[Pawn][e4][Opening] += 20;
	pvt[Pawn][d5][Opening] += 10;
	pvt[Pawn][e5][Opening] += 10;

	/* Knights */
	for(i=0; i<64; i++){
		pvt[Knight][i][Opening] += Knight_Line[file(i)] * Knight_Center_Opening;
		pvt[Knight][i][Opening] += Knight_Line[rank(i)] * Knight_Center_Opening;
		pvt[Knight][i][Opening] += Knight_Rank[rank(i)] * Knight_Rank_Opening;
		pvt[Knight][i][Endgame] += Knight_Line[file(i)] * Knight_Center_Endgame;
		pvt[Knight][i][Endgame] += Knight_Line[rank(i)] * Knight_Center_Endgame;
	}
	/* Back rank */
	for(i=a1; i<=h1; i++) pvt[Knight][i][Opening] -= Knight_Back_Rank_Opening;
	/* Trapped */
	pvt[Knight][a8][Opening] -= Knight_Trapped;
	pvt[Knight][h8][Opening] -= Knight_Trapped;

	/* Bishops */
	for(i=0; i<64; i++){
		pvt[Bishop][i][Opening] += Bishop_Line[file(i)] * Bishop_Center_Opening;
		pvt[Bishop][i][Opening] += Bishop_Line[rank(i)] * Bishop_Center_Opening;
		pvt[Bishop][i][Endgame] += Bishop_Line[file(i)] * Bishop_Center_Endgame;
		pvt[Bishop][i][Endgame] += Bishop_Line[rank(i)] * Bishop_Center_Endgame;
	}
	/* Back rank */
	for(i=a1; i<=h1; i++) pvt[Bishop][i][Opening] -= Bishop_Back_Rank_Opening;
	/* Main diagonals*/
	for(i=0; i<8; i++){
		j = getSquare(i,i);
		pvt[Bishop][j][Opening] += Bishop_Diagonal_Opening;
		pvt[Bishop][opposite(j)][Opening] += Bishop_Diagonal_Opening;
	}

	/* Rooks */
	for(i=0; i<64; i++) pvt[Rook][i][Opening] += Rook_File[file(i)] * Rook_File_Opening;

	/* Queens */
	for(i=0; i<64; i++){
		pvt[Queen][i][Opening] += Queen_Line[file(i)] * Queen_Center_Opening;
		pvt[Queen][i][Opening] += Queen_Line[rank(i)] * Queen_Center_Opening;
		pvt[Queen][i][Endgame] += Queen_Line[file(i)] * Queen_Center_Endgame;
		pvt[Queen][i][Endgame] += Queen_Line[rank(i)] * Queen_Center_Endgame;
	}
	/* Back rank */
	for(i=a1; i<=h1; i++) pvt[Queen][i][Opening] -= Queen_Back_Rank_Opening;

	/* Kings */
	for(i=0; i< 64; i++){
		pvt[King][i][Opening] += King_File[file(i)] * King_File_Opening;
		pvt[King][i][Opening] += King_Rank[rank(i)] * King_Rank_Opening;
		pvt[King][i][Endgame] += King_Line[file(i)] * King_Center_Endgame;
		pvt[King][i][Endgame] += King_Line[rank(i)] * King_Center_Endgame;
	}

	/* Symmetry copy for black */
	for (i = 0; i < 14; i += 2) {
		for (j = 0; j < 64; j++) {
			for (k= 0; k < 2; k++) {
				pvt[i+1][j][k] = -pvt[i][opposite(j)][k];
         		}
      		}
   	}

	/* Print pvt */
	/*for (k= 0; k < 2; k++) {
		for (i = 0; i < 14; i ++) {
			g_print("%c:\n",pieceToChar[i]);
			for (j = 0; j < 64; j++) {
				if(j%8==0) g_print("\n");
				g_print("%d ",pvt[i][j][k]);
			}
			g_print("\n");
      		}
		g_print("\n");
   	}*/
}

void initAttack(void){
	gint s; /* square, bit, diagonal, file, rank*/
	guint64 mask;
	for (s=0;s<64;s++) { /* pre-compute attack masks for w,b pawns: two side squares */
		mask = 1ull<<s;
		attackPawn[White][s]  = shiftUpLeft(mask)   | shiftUpRight(mask);
		attackPawn[Black][s]  = shiftDownLeft(mask) | shiftDownRight(mask);
	}
	for(s=0; s<64; s++){ /** pre-compute 8 knight's attack squares */
		mask = 1ull<<s;
		attackKnight[s] = shiftLeft(shift2Up(mask))    | shiftRight(shift2Up(mask))   |
				  shiftLeft(shift2Down(mask))  | shiftRight(shift2Down(mask)) |
				  shift2Left(shiftUp(mask))    | shift2Right(shiftUp(mask))   |
				  shift2Left(shiftDown(mask))  | shift2Right(shiftDown(mask));
	}
	for (s=0;s<64;s++) {
	 	mask = 1ull<<s;
	 	attackKing[s] = shiftLeft(mask) | shiftRight(mask) |
				shiftUp(mask)   | shiftDown(mask)  |
				shiftUpLeft(mask) | shiftUpRight(mask) |
				shiftDownLeft(mask) | shiftDownRight(mask);
	}

	for (s=0;s<64;s++) {
		maskBishop[s] = positionAttackBishopAll(s);
		maskRook[s] = positionAttackRookAll(s);
		maskQueen[s] =	positionAttackQueenAll(s);
		maskKnight[s] = attackKnight[s];
	}

	for(s=0; s<64; s++){
		mask = attackKing[s];
		kingSafeDistance1[s] = mask;
		kingSafeDistance2[s] = mask | shiftUpLeft(mask) | shiftUpRight(mask) |
								shiftDownLeft(mask) | shiftDownRight(mask);
	}
}

guint iteratedBitCount(guint n){
    guint count=0;    
    while (n){
        count += n & 0x1u ;    
        n >>= 1 ;
    }
    return count ;
}

void initBitsIn16bits(void){
    guint i ;    
    for (i = 0; i < (0x1u<<16); i++)
        bitsIn16bits[i] = iteratedBitCount(i);
    return ;
}

guint64 initmagicmoves_occ(const int* squares, const int numSquares, const guint64 linocc)
{
	int i;
	guint64 ret=0;
	for(i=0;i<numSquares;i++)
		if(linocc&(((guint64)(1))<<i)) ret|=(((guint64)(1))<<squares[i]);
	return ret;
}

guint64 initmagicmoves_Rmoves(const int square, const guint64 occ){
	guint64 ret=0;
	guint64 bit;
	guint64 rowbits=(((guint64)0xFF)<<(8*(square/8)));
	
	bit=(((guint64)(1))<<square);
	do{
		bit<<=8;
		ret|=bit;
	}while(bit && !(bit&occ));
	bit=(((guint64)(1))<<square);
	do{
		bit>>=8;
		ret|=bit;
	}while(bit && !(bit&occ));
	bit=(((guint64)(1))<<square);
	do{
		bit<<=1;
		if(bit&rowbits) ret|=bit;
		else break;
	}while(!(bit&occ));
	bit=(((guint64)(1))<<square);
	do
	{
		bit>>=1;
		if(bit&rowbits) ret|=bit;
		else break;
	}while(!(bit&occ));
	return ret;
}

guint64 initmagicmoves_Bmoves(const int square, const guint64 occ){
	guint64 ret=0;
	guint64 bit;
	guint64 bit2;
	guint64 rowbits=(((guint64)0xFF)<<(8*(square/8)));
	
	bit=(((guint64)(1))<<square);
	bit2=bit;
	do{
		bit<<=8-1;
		bit2>>=1;
		if(bit2&rowbits) ret|=bit;
		else break;
	}while(bit && !(bit&occ));
	bit=(((guint64)(1))<<square);
	bit2=bit;
	do{
		bit<<=8+1;
		bit2<<=1;
		if(bit2&rowbits) ret|=bit;
		else break;
	}while(bit && !(bit&occ));
	bit=(((guint64)(1))<<square);
	bit2=bit;
	do{
		bit>>=8-1;
		bit2<<=1;
		if(bit2&rowbits) ret|=bit;
		else break;
	}while(bit && !(bit&occ));
	bit=(((guint64)(1))<<square);
	bit2=bit;
	do{
		bit>>=8+1;
		bit2>>=1;
		if(bit2&rowbits) ret|=bit;
		else break;
	}while(bit && !(bit&occ));
	return ret;
}

#define BmagicNOMASK2(square, occupancy) *(magicmoves_b_indices2[square]+(((occupancy)*magicmoves_b_magics[square])>>magicmoves_b_shift[square]))
#define RmagicNOMASK2(square, occupancy) *(magicmoves_r_indices2[square]+(((occupancy)*magicmoves_r_magics[square])>>magicmoves_r_shift[square]))

void initmagicmoves(void){
	int i;

	int initmagicmoves_bitpos64_database[64] = {
	63,  0, 58,  1, 59, 47, 53,  2,
	60, 39, 48, 27, 54, 33, 42,  3,
	61, 51, 37, 40, 49, 18, 28, 20,
	55, 30, 34, 11, 43, 14, 22,  4,
	62, 57, 46, 52, 38, 26, 32, 41,
	50, 36, 17, 19, 29, 10, 13, 21,
	56, 45, 25, 31, 35, 16,  9, 12,
	44, 24, 15,  8, 23,  7,  6,  5
	};

	guint64* magicmoves_b_indices2[64] = {
		magicmovesbdb+4992, magicmovesbdb+2624,  magicmovesbdb+256,  magicmovesbdb+896,
		magicmovesbdb+1280, magicmovesbdb+1664, magicmovesbdb+4800, magicmovesbdb+5120,
		magicmovesbdb+2560, magicmovesbdb+2656,  magicmovesbdb+288,  magicmovesbdb+928,
		magicmovesbdb+1312, magicmovesbdb+1696, magicmovesbdb+4832, magicmovesbdb+4928,
		magicmovesbdb+0,     magicmovesbdb+128,  magicmovesbdb+320,  magicmovesbdb+960,
		magicmovesbdb+1344, magicmovesbdb+1728, magicmovesbdb+2304, magicmovesbdb+2432,
		magicmovesbdb+32,    magicmovesbdb+160,  magicmovesbdb+448, magicmovesbdb+2752,
		magicmovesbdb+3776, magicmovesbdb+1856, magicmovesbdb+2336, magicmovesbdb+2464,
		magicmovesbdb+64,    magicmovesbdb+192,  magicmovesbdb+576, magicmovesbdb+3264,
		magicmovesbdb+4288, magicmovesbdb+1984, magicmovesbdb+2368, magicmovesbdb+2496,
		magicmovesbdb+96,    magicmovesbdb+224,  magicmovesbdb+704, magicmovesbdb+1088,
		magicmovesbdb+1472, magicmovesbdb+2112, magicmovesbdb+2400, magicmovesbdb+2528,
		magicmovesbdb+2592, magicmovesbdb+2688,  magicmovesbdb+832, magicmovesbdb+1216,
		magicmovesbdb+1600, magicmovesbdb+2240, magicmovesbdb+4864, magicmovesbdb+4960,
		magicmovesbdb+5056, magicmovesbdb+2720,  magicmovesbdb+864, magicmovesbdb+1248,
		magicmovesbdb+1632, magicmovesbdb+2272, magicmovesbdb+4896, magicmovesbdb+5184
	};
	guint64* magicmoves_r_indices2[64] = {
		magicmovesrdb+86016, magicmovesrdb+73728, magicmovesrdb+36864, magicmovesrdb+43008,
		magicmovesrdb+47104, magicmovesrdb+51200, magicmovesrdb+77824, magicmovesrdb+94208,
		magicmovesrdb+69632, magicmovesrdb+32768, magicmovesrdb+38912, magicmovesrdb+10240,
		magicmovesrdb+14336, magicmovesrdb+53248, magicmovesrdb+57344, magicmovesrdb+81920,
		magicmovesrdb+24576, magicmovesrdb+33792,  magicmovesrdb+6144, magicmovesrdb+11264,
		magicmovesrdb+15360, magicmovesrdb+18432, magicmovesrdb+58368, magicmovesrdb+61440,
		magicmovesrdb+26624,  magicmovesrdb+4096,  magicmovesrdb+7168,     magicmovesrdb+0,
		magicmovesrdb+2048,  magicmovesrdb+19456, magicmovesrdb+22528, magicmovesrdb+63488,
		magicmovesrdb+28672,  magicmovesrdb+5120,  magicmovesrdb+8192,  magicmovesrdb+1024,
		magicmovesrdb+3072,  magicmovesrdb+20480, magicmovesrdb+23552, magicmovesrdb+65536,
		magicmovesrdb+30720, magicmovesrdb+34816,  magicmovesrdb+9216, magicmovesrdb+12288,
		magicmovesrdb+16384, magicmovesrdb+21504, magicmovesrdb+59392, magicmovesrdb+67584,
		magicmovesrdb+71680, magicmovesrdb+35840, magicmovesrdb+39936, magicmovesrdb+13312,
		magicmovesrdb+17408, magicmovesrdb+54272, magicmovesrdb+60416, magicmovesrdb+83968,
		magicmovesrdb+90112, magicmovesrdb+75776, magicmovesrdb+40960, magicmovesrdb+45056,
		magicmovesrdb+49152, magicmovesrdb+55296, magicmovesrdb+79872, magicmovesrdb+98304
	};

	for(i=0;i<64;i++){
		int squares[64];
		int numsquares=0;
		guint64 temp=magicmoves_b_mask[i];
		while(temp)
		{
			guint64 bit=temp&-temp;
			squares[numsquares++]=initmagicmoves_bitpos64_database[(bit*C64(0x07EDD5E59A4E28C2))>>58];
			temp^=bit;
		}
		for(temp=0;temp<(((guint64)(1))<<numsquares);temp++)
		{
		 guint64 tempocc=initmagicmoves_occ(squares,numsquares,temp);
				BmagicNOMASK2(i,tempocc)=initmagicmoves_Bmoves(i,tempocc);
		}
	}
	for(i=0;i<64;i++)
	{
		int squares[64];
		int numsquares=0;
		guint64 temp=magicmoves_r_mask[i];
		while(temp)
		{
			guint64 bit=temp&-temp;
			squares[numsquares++]=initmagicmoves_bitpos64_database[(bit*C64(0x07EDD5E59A4E28C2))>>58];
			temp^=bit;
		}
		for(temp=0;temp<(((guint64)(1))<<numsquares);temp++)
		{
			guint64 tempocc=initmagicmoves_occ(squares,numsquares,temp);
				RmagicNOMASK2(i,tempocc)=initmagicmoves_Rmoves(i,tempocc);
		}
	}
}
